<?php

namespace Drupal\webt\Controller;

use Drupal\webt\translation_engines\etranslation\EtranslationUtils;
use Drupal\Component\Utility\Html;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller that handles eTranslation async responses.
 */
class EtranslationCallbackController {

	/**
	 * Receive & process async translation response from eTranslation.
	 * If this is a translaton for an entity and timeout has been reached before, also parse and save translation.
	 *
	 * @return Symfony\Component\HttpFoundation\Response
	 */
	public function onResponse() {
		$request_id         = Html::escape( $_REQUEST['request-id'] );
		$translated_content = base64_decode( file_get_contents( 'php://input' ) );
		$key                = EtranslationUtils::get_cache_key( $request_id );
		\Drupal::logger( WEBT_LOGGER )->info( "Received response from eTranslation ($request_id)" );
		if ( \Drupal::cache()->get( $key ) && \Drupal::cache()->get( $key )->data === EtranslationUtils::$timeout_value ) {
			\Drupal::logger( WEBT_LOGGER )->info( "Received late translation ($request_id)" );
			$entity_id = Html::escape( $_REQUEST['external-reference'] );
			if ( $entity_id ) {
				$entity = $this->loadEntityById( $entity_id );
				if ( $entity ) {
					$this->processEntityTranslation( $entity, $translated_content, Html::escape( $_REQUEST['target-language'] ) );
				}
			}
		} else {
			\Drupal::cache()->set( $key, $translated_content );
		}
		return new Response();
	}

	/**
	 * Process async error response from eTranslation - log error and mark translation as completed with error.
	 *
	 * @return Symfony\Component\HttpFoundation\Response
	 */
	public function onError() {
		$request_id = Html::escape( $_REQUEST['request-id'] );
		\Drupal::logger( WEBT_LOGGER )->error( "eTranslation response error ($request_id). Please make sure eTranslation API is working as expected." );
		\Drupal::cache()->set( EtranslationUtils::get_cache_key( $request_id ), EtranslationUtils::$error_value );
		return new Response();
	}

	/**
	 * Parse translation response and insert entity translation.
	 *
	 * @param [type] $entity Original entity to translate.
	 * @param [type] $translated_content Translation received from eTranslation service.
	 * @param [type] $target_language Language of translation received.
	 * @return void
	 */
	private function processEntityTranslation( $entity, $translated_content, $target_language ) {
		$source_langcode     = $entity->language()->getId();
		$target_langcode     = $this->getLangcodeByLanguage( $target_language );
		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$translation_service = \Drupal::service( 'webt.translation_service.etranslation' );
		$xml_translations    = $translation_service->etranslation_response_to_translation_array( $translated_content );

		$translatable_fields = $translation_manager->get_translatable_fields( $entity );
		$original_values     = $translation_manager->extract_translatable_values( $entity, $translatable_fields );
		if ( count( $original_values ) === count( $xml_translations ) ) {
			$translations = $translation_service->decode_xml_translations( $original_values, $xml_translations );
			$translation_manager->translate_entity( $entity, $source_langcode, $target_langcode, $translations );
		}
	}

	/**
	 * Retrieve entity by ID
	 *
	 * @param [type] $id Identifier of entity.
	 * @return \Drupal\Core\Entity\Entity or null if enitity was not found
	 */
	private function loadEntityById( $id ) {
		$entity_type_manager = \Drupal::service( 'entity_type.manager' );
		foreach ( $entity_type_manager->getDefinitions() as $entity_type_id => $entity_type ) {
			if ( $entity_type->entityClassImplements( '\Drupal\Core\Entity\EntityInterface' ) ) {
				$storage = $entity_type_manager->getStorage( $entity_type_id );
				$entity  = $storage->load( $id );
				if ( $entity ) {
					return $entity;
				}
			}
		}
		\Drupal::logger( WEBT_LOGGER )->error( "Could not load entity by ID ($id). Please make sure such entity exists!" );
		return null;
	}

	/**
	 * In list of languages added to the website find the first language code for given language.
	 *
	 * @param string $language Requested language e.g. 'en'.
	 * @return string or null if langcode not found.
	 */
	private function getLangcodeByLanguage( $language ) {
		$languages = \Drupal::languageManager()->getLanguages();
		$language_lowercase = strtolower( $language );

		foreach ( $languages as $langcode => $value ) {
			if ( str_starts_with( $langcode, $language_lowercase ) ) {
				return $langcode;
			}
		}
		\Drupal::logger( WEBT_LOGGER )->error( "Could not find langcode for language '$language_lowercase'. Please check if the language code of this language is correct!" );
		return null;
	}
}
