<?php

namespace Drupal\webt\Controller;

use Symfony\Component\HttpFoundation\Response;

/**
 * Controller that handles ajax pre-translation progress requests from frontend.
 */
class PretranslationProgressController {

	/**
	 * Cache key that stores pre-translation progress value.
	 *
	 * @var string
	 */
	public static $progress_key = 'webt_translation_progress';

	/**
	 * Retrieves translation progress message and percentage from Drupal cache.
	 *
	 * @return Symfony\Component\HttpFoundation\Response
	 */
	public function getTranslationProgress() {
		$response = null;
		$value    = \Drupal::cache()->get( self::$progress_key );
		if ( $value && $value->data ) {
			$response = $value->data;
		} else {
			$empty_response             = new \stdClass();
			$empty_response->message    = '';
			$empty_response->percentage = 0;
			$response                   = $empty_response;
		}
		return new Response( json_encode( $response ) );
	}

	/**
	 * Removes translation progress message and percentage from Drupal cache.
	 *
	 * @return Symfony\Component\HttpFoundation\Response
	 */
	public function cancelTranslation() {
		\Drupal::logger( WEBT_LOGGER )->error( 'Client page was closed, cancelling pre-translation...' );
		self::updateProgress( null, null, true );
		return new Response();
	}

	/**
	 * Add error notification (to be called on ajax error)
	 *
	 * @return Symfony\Component\HttpFoundation\Response
	 */
	public function ajaxErrorHandler() {
		$logs = \Drupal\Core\Link::fromTextAndUrl( 'logs', \Drupal\Core\Url::fromRoute('dblog.overview') );
		\Drupal::messenger()->addError( t( 'Backend translation process was aborted! Check @logs for errors or try increasing website\'s PHP max_execution_time.', array( '@logs' => $logs->toString() ) ) );
		return new Response();
	}

	/**
	 * Update progress bar message
	 *
	 * @param string $message Message to be shown below progress bar.
	 * @return void
	 */
	public static function updateTranslationMessage( $message ) {
		self::updateProgress( null, $message );
	}

	/**
	 * Update progress bar percentage
	 *
	 * @param float $percentage Percentage value to be shown in progress bar.
	 * @return void
	 */
	public static function updateTranslationPercentage( $percentage ) {
		self::updateProgress( $percentage );
	}

	/**
	 * Delete translation progress information from cache.
	 *
	 * @return void
	 */
	public static function clearTranslationProgress() {
		\Drupal::cache()->delete( self::$progress_key );
	}

	/**
	 * Retrieve translation progress information from cache.
	 *
	 * @return stdClass
	 */
	public static function getValue() {
		return \Drupal::cache()->get( self::$progress_key );
	}

	/**
	 * Update translation progress information in cache.
	 *
	 * @param float $percentage New percentage.
	 * @param string  $message New message.
	 * @param boolean $aborted Is this translation process aborted.
	 * @return void
	 */
	private static function updateProgress( $percentage = null, $message = null, $aborted = false ) {
		$value  = \Drupal::cache()->get( self::$progress_key );
		$update = null;
		if ( $value && $value->data ) {
			$update = $value->data;
		} else {
			$update             = new \stdClass();
			$update->message    = '';
			$update->percentage = 0;
			$update->aborted    = false;
			$update->request_id = uniqid();
		}
		if ( $percentage !== null ) {
			$update->percentage = round( $percentage );
		}
		if ( $message ) {
			$update->message = $message;
		}
		if ( $aborted ) {
			$update->aborted = $aborted;
		}
		\Drupal::cache()->set( self::$progress_key, $update );
	}
}
