<?php

namespace Drupal\webt\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Url;
use Drupal\webt\Controller\PretranslationProgressController;
use Drupal\webt\Model\ProcessingResponse;
use Drupal\webt\Model\StringType;

/**
 * Pre-translation tab form.
 */
class PretranslationForm extends FormBase {

	/**
	 * Drupal form id.
	 *
	 * @return string
	 */
	public function getFormId() {
		return 'PretranslationForm';
	}

	/**
	 * Adds fields and libraries to the form.
	 *
	 * @param array              $form Form.
	 * @param FormStateInterface $form_state Form state.
	 * @return Drupal\Core\Form\FormInterface
	 */
	public function buildForm( array $form, FormStateInterface $form_state ) {
		$selected_types = array_filter(
			StringType::get_all_types(),
			function( $type ) {
				return $this->is_string_type_selected( $type );
			},
		);
		$selected_types = array_map(
			function( $type ) {
				return StringType::get_display_name( $type );
			},
			$selected_types
		);

		$form['description'] = array(
			'#type'   => 'markup',
			'#markup' => '<div class="form-item__description">' . t( 'Follow translation progress and translate existing content. New content will be translated automatically.' ) . '</div>',
		);

		$form['note'] = array(
			'#type'   => 'markup',
			'#markup' => '<div class="form-item__description"><strong>' . t( 'Note' ) . ': </strong>' . t( 'Translating different languages or content types separately can be more effiecient.' ) . '</div>',
		);

		$form['table'] = $this->getTranslationProgressTable();

		$translate_button = array(
			'#type'  => 'submit',
			'#value' => t( 'Translate' ),
			'#ajax'  => array(
				'callback' => '::translateItems',
				'event'    => 'click',
				'progress' => array(
					'type'     => 'bar',
					'message'  => 'Translating...',
					'url'      => Url::fromRoute( 'webt.pretranslation.progress' ),
					'interval' => '500',
				),
			),
		);

		$confirm_deletion = t( 'Are you sure you want to delete translations of selected languages?' );
		$delete_button = array(
			'#type'       => 'submit',
			'#value'      => t( 'Delete translations' ),
			'#submit'     => array( '::deleteItems' ),
			'#attributes' => array(
				'title'   => t( 'Delete translations of selected languages and types' ),
				'onclick' => "return confirm('$confirm_deletion');",
			),
		);

		$form['button_wrapper'] = array(
			'#type'            => 'container',
			'#attributes'      => array( 'class' => array( 'text-align-right' ) ),
			'delete'           => $delete_button,
			'translate_button' => $translate_button,
		);

		$form['advanced'] = $this->getAdvancedOptions();

		$form['#attached']['library'][]                        = 'core/drupal.ajax';
		$form['#attached']['library'][]                        = 'webt/pretranslation_tab.style';
		$form['#attached']['library'][]                        = 'webt/pretranslation_tab.js';
		$form['#attached']['drupalSettings']['selected_types'] = $selected_types;

		return $form;
	}

	/**
	 * Creates Advanced option section with fields.
	 *
	 * @return array
	 */
	private function getAdvancedOptions() {
		$advanced_options = array(
			'#type'  => 'details',
			'#title' => t( 'Content types' ),
			'#open'  => false,
			'#tree'  => true,
		);

		$advanced_options['description'] = array(
			'#type'   => 'markup',
			'#markup' => '<div class="form-item__description">' . t( 'Selected content types to be translated/deleted.' ) . '</div>',
		);

		$advanced_options[ StringType::NODE ] = array(
			'#type'          => 'checkbox',
			'#title'         => t( 'Nodes' ),
			'#default_value' => $this->is_string_type_selected( StringType::NODE ),
			'#description'   => t( 'Translate untranslated node type elements (articles, pages)' ),
		);

		$advanced_options[ StringType::COMMENT ] = array(
			'#type'          => 'checkbox',
			'#title'         => t( 'Comments' ),
			'#default_value' => $this->is_string_type_selected( StringType::COMMENT ),
			'#description'   => t( 'Translate untranslated comments' ),
		);

		$advanced_options[ StringType::TAG ] = array(
			'#type'          => 'checkbox',
			'#title'         => t( 'Tags' ),
			'#default_value' => $this->is_string_type_selected( StringType::TAG ),
			'#description'   => t( 'Translate untranslated taxonomy terms (tags)' ),
		);

		$advanced_options[ StringType::CONFIG ] = array(
			'#type'          => 'checkbox',
			'#title'         => t( 'Configuration' ),
			'#default_value' => $this->is_string_type_selected( StringType::CONFIG ),
			'#description'   => t( 'Translate untranslated configuration strings (site name, role names, blocks, etc.)' ),
		);

		$advanced_options[ StringType::UI ] = array(
			'#type'          => 'checkbox',
			'#title'         => t( 'UI elements' ),
			'#default_value' => $this->is_string_type_selected( StringType::UI ),
			'#description'   => t( 'Translate untranslated UI elements' ),
		);

		$advanced_options['save_wrapper'] = array(
			'#type'       => 'container',
			'#attributes' => array( 'class' => array( 'text-align-right' ) ),
			'save'        => array(
				'#type'       => 'submit',
				'#value'      => t( 'Save' ),
				'#attributes' => array( 'title' => t( 'Save affected content type selection' ) ),
			),
		);

		// Checks if specific content types are not translatable due to website configuration.
		$this->checkContentTypeTranslationEnabled( StringType::NODE, array( 'article', 'page' ) );
		$this->checkContentTypeTranslationEnabled( StringType::COMMENT, array( 'comment' ) );
		$this->checkContentTypeTranslationEnabled( StringType::TAG, array( 'tags' ) );

		return $advanced_options;
	}

	/**
	 * Updates translatable content type selection on Advanced option save button click.
	 *
	 * @param array              $form Form.
	 * @param FormStateInterface $form_state Form state.
	 * @return void
	 */
	public function submitForm( array &$form, FormStateInterface $form_state ) {
		\Drupal::messenger()->deleteAll();
		$config = \Drupal::getContainer()->get( 'config.factory' )->getEditable( 'webt.settings' );
		$types  = StringType::get_all_types();
		foreach ( $types as $type ) {
			$selected = $form_state->getValue( 'advanced' )[ $type ];
			$config->set( 'translate_' . $type, $selected );
		}
		$config->save();
		\Drupal::messenger()->addStatus( t( 'Affected content type selection saved' ) );
	}

	/**
	 * Checks if translation is enabled in website settings for given content type.
	 * If content type is marked as translatable in this module but not in website config, warning is added.
	 *
	 * @param string   $entity_type_id Content type ID.
	 * @param string[] $bundle_names Array of properties to check for given $entity_type_id.
	 * @return boolean
	 */
	public function checkContentTypeTranslationEnabled( $entity_type_id, $bundle_names ) {
		$content_tranlation_manager = \Drupal::service( 'content_translation.manager' );
		foreach ( $bundle_names as $bundle_name ) {
			if ( $content_tranlation_manager->isEnabled( $entity_type_id, $bundle_name ) ) {
				return true;
			}
		}
		if ( $this->is_string_type_selected( $entity_type_id ) ) {
			$settings = \Drupal\Core\Link::fromTextAndUrl( t( 'Content language settings' ), \Drupal\Core\Url::fromRoute( 'language.content_settings_page' ) );
			\Drupal::messenger()->addWarning( t( StringType::get_display_name( $entity_type_id ) . ' are not set as translatable in @settings!', array( '@settings' => $settings->toString() ) ) );

		}

		return false;
	}

	/**
	 * Pre-translate website to selected languages and content types.
	 *
	 * @param array              $form Form.
	 * @param FormStateInterface $form_state Form state.
	 * @return Drupal\Core\Ajax\AjaxResponse
	 */
	public function translateItems( array &$form, FormStateInterface $form_state ) {
		$response = new AjaxResponse();
		\Drupal::messenger()->deleteAll();

		if ( ! $this->validate_pretranslation_request() ) {
			return $response;
		}

		PretranslationProgressController::clearTranslationProgress();
		PretranslationProgressController::updateTranslationMessage( 'Translation started' );
		PretranslationProgressController::updateTranslationPercentage( 0 );

		$selected_langcodes = array_filter( $form_state->getValue( 'table' ) );
		if ( empty( $selected_langcodes ) ) {
			\Drupal::messenger()->addWarning( t( 'Please select at least one target language' ) );
			return $response;
		}
		$selected_langcodes = $this->filter_supported_languages( $selected_langcodes );

		if ( $this->is_string_type_selected( StringType::TAG ) ) {
			$this->translateEntities( StringType::TAG, $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::NODE ) ) {
			$this->translateEntities( StringType::NODE, $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::COMMENT ) ) {
			$this->translateEntities( StringType::COMMENT, $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::CONFIG ) ) {
			$this->translateConfigurationStrings( $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::UI ) ) {
			$this->translateUserInterfaceElements( $selected_langcodes );
		}
		drupal_flush_all_caches();

		\Drupal::logger( WEBT_LOGGER )->info( 'Website translation finished' );
		return $response;
	}

	/**
	 * Translate all entities of give type.
	 *
	 * @param string   $entity_type Entity type.
	 * @param string[] $langcodes Target language codes.
	 * @return void
	 */
	public function translateEntities( $entity_type, $langcodes ) {
		\Drupal::logger( WEBT_LOGGER )->debug( "Translating entities of type: '$entity_type'" );

		$entities = \Drupal::entityTypeManager()
			->getStorage( $entity_type )
			->loadMultiple();

		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$response            = $translation_manager->translate_entities( $entities, $entity_type, $langcodes );
		$this->register_translation_results( $response, $entity_type );
	}

	/**
	 * Translate UI strings.
	 *
	 * @param string[] $langcodes Target language codes.
	 * @return void
	 */
	public function translateUserInterfaceElements( $langcodes ) {
		\Drupal::logger( WEBT_LOGGER )->debug( 'Translating UI elements' );
		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$source_language     = $translation_manager->get_default_language_code();

		$results = array( ProcessingResponse::$already_processed );
		foreach ( $langcodes as $langcode ) {
			if ( $langcode === $source_language ) {
				continue;
			}
			$untranslated_strings = $translation_manager->get_untranslated_ui_strings( $langcode );
			$results[]            = $translation_manager->translate_ui_strings( $untranslated_strings, $source_language, $langcode );
		}
		$response = min( $results );
		$this->register_translation_results( $response, StringType::UI );
	}

	/**
	 * Translate configuration strings.
	 *
	 * @param string[] $langcodes Target language codes.
	 * @return void
	 */
	public function translateConfigurationStrings( $langcodes ) {
		\Drupal::logger( WEBT_LOGGER )->debug( 'Translating configurations strings' );
		PretranslationProgressController::updateTranslationPercentage( 0 );

		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$source_language     = $translation_manager->get_default_language_code();

		$results = array( ProcessingResponse::$already_processed );
		foreach ( $langcodes as $langcode ) {
			if ( $langcode === $source_language ) {
				continue;
			}
			$strings   = $translation_manager->get_untranslated_config_strings( $langcode );
			$results[] = $translation_manager->translate_config_strings( $strings, $langcode );
		}
		$response = min( $results );
		$this->register_translation_results( $response, StringType::CONFIG );
	}

	/**
	 * Delete translations for selected languages and content types.
	 *
	 * @param array              $form Form.
	 * @param FormStateInterface $form_state Form state.
	 * @return void
	 */
	public function deleteItems( array &$form, FormStateInterface $form_state ) {
		$translation_manager = \Drupal::service( 'webt.translation_manager' );

		$selected_langcodes  = array_filter( $form_state->getValue( 'table' ) );
		if ( empty( $selected_langcodes ) ) {
			\Drupal::messenger()->addWarning( t( 'Please select at least one target language' ) );
			return;
		}

		if ( $this->is_string_type_selected( StringType::TAG ) ) {
			$this->deleteEntityTranslations( StringType::TAG, $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::NODE ) ) {
			$this->deleteEntityTranslations( StringType::NODE, $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::COMMENT ) ) {
			$this->deleteEntityTranslations( StringType::COMMENT, $selected_langcodes );
		}
		if ( $this->is_string_type_selected( StringType::CONFIG ) ) {
			\Drupal::logger( WEBT_LOGGER )->debug( 'Deleting configuration translations' );
			$response = $translation_manager->delete_config_translations( $selected_langcodes );
			$this->register_deletion_results( $response, StringType::CONFIG );
		}
		if ( $this->is_string_type_selected( StringType::UI ) ) {
			\Drupal::logger( WEBT_LOGGER )->debug( 'Deleting UI string translations' );
			$response = $translation_manager->delete_ui_translations( $selected_langcodes );
			$this->register_deletion_results( $response, StringType::UI );
		}
	}

	/**
	 * Delete entity translations for given languages.
	 *
	 * @param string   $entity_type Entity type ID.
	 * @param string[] $langcodes Target language codes.
	 * @return void
	 */
	public function deleteEntityTranslations( $entity_type, $langcodes ) {
		\Drupal::logger( WEBT_LOGGER )->debug( "Deleting entity translations of type: '$entity_type'" );

		$entities = \Drupal::entityTypeManager()
			->getStorage( $entity_type )
			->loadMultiple();

		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$response            = $translation_manager->delete_translations_of_entities( $entities, $langcodes );
		$this->register_deletion_results( $response, $entity_type );
	}

	/**
	 * Build translation progress table.
	 *
	 * @return array
	 */
	public function getTranslationProgressTable() {
		$types                = StringType::get_all_types();
		$translation_manager  = \Drupal::service( 'webt.translation_manager' );
		$langcodes            = \Drupal::languageManager()->getLanguages();
		$langcodes_list       = array_keys( $langcodes );
		$translation_progress = array();

		foreach ( $langcodes_list as $langcode ) {
			$translation_progress[ $langcode ] = array();
			foreach ( $types as $type ) {
				$translation_progress[ $langcode ][ $type ] = $translation_manager->get_translation_progress_percents( $type, $langcode );
			}
		}

		$rows = array();
		foreach ( $translation_progress as $langcode => $value ) {
			$language     = t( $langcodes[ $langcode ]->getName() );
			$row          = array( $language );
			$row['title'] = array(
				'data' => array(
					'#title' => $language,
				),
			);
			foreach ( $value as $progress ) {
				$row[] = round( $progress, 2 ) . '%';
			}
			$rows[ $langcode ] = $row;
		}

		$display_types = array_map(
			function( $key ) {
				return StringType::get_display_name( $key );
			},
			$types
		);

		$header = array( t( 'Language' ) ) + $display_types;
		return array(
			'#type'    => 'tableselect',
			'#header'  => array_values( $header ),
			'#options' => $rows,
		);
	}

	/**
	 * Add translation finished notification based on given status response.
	 *
	 * @param integer $response Pre-translation process response.
	 * @param string  $display_type Type of strings that were translated.
	 * @return void
	 */
	public function register_translation_results( $response, $display_type ) {
		switch ( $response ) {
			case ProcessingResponse::$fully_processed:
				\Drupal::messenger()->addStatus( t( "Type '$display_type' strings translated successfully!" ) );
				return;
			case ProcessingResponse::$partially_processed:
				$logs = \Drupal\Core\Link::fromTextAndUrl( 'logs', \Drupal\Core\Url::fromRoute( 'dblog.overview' ) );
				\Drupal::messenger()->addWarning( t( "Type '$display_type' strings translated partially. Try decreasing MT engine request size (Settings->Advanced->Max characters per single request) or check @logs for errors.", array( '@logs' => $logs->toString() ) ) );
				return;
			case ProcessingResponse::$processing_error:
				$logs = \Drupal\Core\Link::fromTextAndUrl( 'logs', \Drupal\Core\Url::fromRoute( 'dblog.overview' ) );
				\Drupal::messenger()->addError( t( "Type '$display_type' strings were not translated. Check your machine-translation configuration or @logs.", array( '@logs' => $logs->toString() ) ) );
				return;
			case ProcessingResponse::$already_processed:
				return;
			default:
				\Drupal::logger( WEBT_LOGGER )->error( "Invalid translation response '$response'. Please contact the developers of this module!" );
				return;
		}
	}

	/**
	 * Add translations deleted notification based on given status response.
	 *
	 * @param integer $response Deletion process response.
	 * @param string  $display_type Type of translations that were deleted.
	 * @return void
	 */
	public function register_deletion_results( $response, $display_type ) {
		switch ( $response ) {
			case ProcessingResponse::$fully_processed:
				\Drupal::messenger()->addStatus( t( "Type '$display_type' strings deleted successfully!" ) );
				return;
			default:
				\Drupal::logger( WEBT_LOGGER )->error( "Invalid deletion response '$response'. Please contact the developers of this module!" );
				return;
		}
	}

	/**
	 * Checks if translation is enabled in module config for given string type.
	 *
	 * @param string $type Translatable object type.
	 * @return boolean
	 */
	private function is_string_type_selected( $type ) {
		$config = $this->config( 'webt.settings' );
		$value = $config->get( 'translate_' . $type );
		return $value === null || $value;
	}

	/**
	 * Filters given language code list and returns only the language codes supported by MT provider.
	 *
	 * @param string[] $selected_langcodes Target language codes.
	 * @return array
	 */
	private function filter_supported_languages( $selected_langcodes ) {
		$supported_languages = array();
		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$translation_service = $translation_manager->translation_service;
		foreach ( $selected_langcodes as $langcode ) {
			$language_manager = \Drupal::languageManager();
			$language = $language_manager->getLanguageName( $langcode );
			if ( ! $translation_service->is_language_supported( $langcode ) ) {
				\Drupal::messenger()->addWarning( t( 'Language is not supported by MT provider: @language', array( '@language' => t( $language ) ) ) );
			} else {
				$supported_languages[] = $langcode;
			}
		}
		return $supported_languages;
	}

	/**
	 * Check if pre-translation request is valid.
	 *
	 * @return boolean
	 */
	private function validate_pretranslation_request() {
		$translation_manager = \Drupal::service( 'webt.translation_manager' );
		$translation_service = $translation_manager->translation_service;
		if ( ! $translation_service->is_authorized() ) {
			\Drupal::messenger()->addError( t( 'Machine translation provider authorization data is not configured. Please go to Translation provider section and enter your MT provider data!' ) );
			return false;
		}
		return true;
	}
}
