<?php

namespace Drupal\webt\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\webt\translation_engines\etranslation\EtranslationService;
use Drupal\webt\translation_engines\etranslation\EtranslationUtils;
use Drupal\webt\translation_engines\generic\GenericMTService;

/**
 * Machine translation provider setting form.
 */
class TranslationProviderForm extends FormBase {

	/**
	 * Internal id of eTranslation provider.
	 *
	 * @var string
	 */
	private $etranslation_engine_id = 'etranslation';
	/**
	 * Internal id of Custom provider.
	 *
	 * @var string
	 */
	private $generic_mt_engine_id = 'generic';

	/**
	 * Retrieve Drupal form id.
	 *
	 * @return string
	 */
	public function getFormId() {
		return 'TranslationProviderForm';
	}

	/**
	 * Adds fields and libraries to the form.
	 *
	 * @param array              $form Form.
	 * @param FormStateInterface $form_state Form state.
	 * @return Drupal\Core\Form\FormInterface
	 */
	public function buildForm( array $form, FormStateInterface $form_state ) {
		$config = $this->config( 'webt.settings' );

		$selected_mt_engine = $config->get( 'mt_engine' );

		// Conditions of current provider radio choice value.
		$condition_etranslation = array(
			':input[name="mt_engine"]' => array( 'value' => $this->etranslation_engine_id ),
		);

		$condition_generic_mt = array(
			':input[name="mt_engine"]' => array( 'value' => $this->generic_mt_engine_id ),
		);

		$condition_selected_engine = array(
			':input[name="mt_engine"]' => array( 'value' => $selected_mt_engine ),
		);

		// Add authorization section.
		$form['authorization'] = array(
			'#type'                 => 'fieldset',
			'#title'                => $this->t( 'Translation provider settings' ),
			'#collapsible'          => false,
			'#collapsed'            => false,
			'description'           => array(
				'#type'   => 'markup',
				'#markup' => '<div class="form-item__description">' . $this->t( 'Choose machine translation provider to use in website translation.' ) . '</div>',
			),
			'mt_engine'             => array(
				'#type'          => 'radios',
				'#title'         => $this->t( 'Translation provider' ),
				'#options'       => array(
					$this->etranslation_engine_id => $this->t( 'eTranslation' ),
					$this->generic_mt_engine_id   => $this->t( 'Custom provider' ),
				),
				'#default_value' => $selected_mt_engine ? $selected_mt_engine : $this->etranslation_engine_id,
			),
			'etranslation_appname'  => array(
				'#type'          => 'textfield',
				'#title'         => $this->t( 'Application name' ),
				'#default_value' => $config->get( 'etranslation_appname' ),
				'#states'        => array(
					'visible'  => $condition_etranslation,
					'required' => $condition_etranslation,
					'disabled' => $condition_generic_mt,
				),
			),
			'etranslation_password' => array(
				'#type'       => 'password',
				'#title'      => $this->t( 'Password' ),
				'#attributes' => array( 'value' => $config->get( 'etranslation_password' ) ),
				'#states'     => array(
					'visible'  => $condition_etranslation,
					'required' => $condition_etranslation,
					'disabled' => $condition_generic_mt,
				),
			),
			'etranslation_help'     => array(
				'#type'       => 'container',
				'#attributes' => array( 'class' => array( 'form-item__description' ) ),
				'#states'     => array(
					'visible' => $condition_etranslation,
				),
				'help_text'   => array(
					'#type'   => 'markup',
					'#markup' => new FormattableMarkup(
						t( "Set up an account for <a target='_blank' href='@link'>eTranslation</a>" ),
						array(
							'@link' => 'https://website-translation.language-tools.ec.europa.eu/automated-translation_en',
						)
					),
				),
			),
			'mt_api_url'            => array(
				'#type'          => 'textfield',
				'#title'         => $this->t( 'Base URL' ),
				'#default_value' => $config->get( 'mt_api_url' ),
				'#placeholder'   => 'https://example.com/api',
				'#states'        => array(
					'visible'  => $condition_generic_mt,
					'required' => $condition_generic_mt,
					'disabled' => $condition_etranslation,
				),
			),
			'mt_api_key'            => array(
				'#type'          => 'textfield',
				'#title'         => $this->t( 'API key' ),
				'#default_value' => $config->get( 'mt_api_key' ),
				'#states'        => array(
					'visible'  => $condition_generic_mt,
					'required' => $condition_generic_mt,
					'disabled' => $condition_etranslation,
				),
			),
			'mt_api_help'           => array(
				'#type'       => 'container',
				'#attributes' => array( 'class' => array( 'form-item__description' ) ),
				'#states'     => array(
					'visible' => $condition_generic_mt,
				),
				'help_text'   => array(
					'#type'   => 'markup',
					'#markup' => new FormattableMarkup(
						t( "Get <a target='_blank' href='@link'>MT provider access</a>" ),
						array(
							'@link' => 'https://website-translation.language-tools.ec.europa.eu/automated-translation_en',
						)
					),
				),
			),
			'button_wrapper'        => array(
				'#type'       => 'container',
				'#attributes' => array( 'class' => array( 'text-align-right' ) ),
				'submit_auth' => array(
					'#type'   => 'submit',
					'#value'  => $this->t( 'Save' ),
					'#name'   => 'submit_auth',
					'#submit' => array( '::submitForm' ),
				),
			),
		);

		// Add configuration section if auth data is saved.
		if ( $this->machineTranslationConfigured() ) {
			$form['configuration'] = array(
				'#type'        => 'fieldset',
				'#title'       => $this->t( 'Configuration' ),
				'#collapsible' => false,
				'#collapsed'   => false,
				'description'  => array(
					'#type'   => 'markup',
					'#markup' => '<div class="form-item__description">' . $this->t( 'Select MT engines to use in translation and configure advanced settings (optional).' ) . '</div>',
				),
				'mt_engines'   => array(
					'#type'        => 'fieldset',
					'#title'       => $this->t( 'Machine translation engines' ),
					'#collapsible' => false,
					'#collapsed'   => false,
					'fields'       => $this->getMtEngineFields(),
				),
				'#states'      => array(
					'visible' => $condition_selected_engine,
				),
			);

			$form['configuration']['etranslation_advanced'] = array(
				'#type'                          => 'details',
				'#title'                         => $this->t( 'Advanced' ),
				'#open'                          => false,
				'#tree'                          => true,
				'#states'                        => array(
					'visible'  => $condition_etranslation,
					'required' => $condition_etranslation,
					'disabled' => $condition_generic_mt,
				),
				'etranslation_timeout'           => array(
					'#type'          => 'number',
					'#title'         => $this->t( 'eTranslation Timeout' ),
					'#description'   => $this->t( 'Max time in seconds to wait for eTranslation to translate new/updated entities on save (per language). Infinite if zero. Does not affect pre-translation.' ),
					'#step'          => '0.1',
					'#default_value' => $config->get( 'etranslation_timeout' ) ? $config->get( 'etranslation_timeout' ) : EtranslationUtils::$etranslation_timeout_default,
				),
				'etranslation_chars_per_request' => array(
					'#type'          => 'number',
					'#title'         => $this->t( 'Max characters per single request' ),
					'#description'   => $this->t( 'Affects how many/big requests are sent to MT provider, which can influence translation speed.' ),
					'#step'          => '100',
					'#default_value' => $config->get( 'etranslation_chars_per_request' ) ? $config->get( 'etranslation_chars_per_request' ) : EtranslationService::$default_chars_per_request,
				),
			);

			$form['configuration']['mt_advanced'] = array(
				'#type'                => 'details',
				'#title'               => $this->t( 'Advanced' ),
				'#open'                => false,
				'#tree'                => true,
				'#states'              => array(
					'visible'  => $condition_generic_mt,
					'required' => $condition_generic_mt,
					'disabled' => $condition_etranslation,
				),
				'mt_chars_per_request' => array(
					'#type'          => 'number',
					'#title'         => $this->t( 'Max characters per single request' ),
					'#description'   => $this->t( 'Affects how many/big requests are sent to MT provider, which can influence translation speed.' ),
					'#step'          => '100',
					'#default_value' => $config->get( 'mt_chars_per_request' ) ? $config->get( 'mt_chars_per_request' ) : GenericMTService::$default_chars_per_request,
				),
			);

			$form['configuration']['button_wrapper'] = array(
				'#type'         => 'container',
				'#attributes'   => array( 'class' => array( 'text-align-right' ) ),
				'submit_config' => array(
					'#type'   => 'submit',
					'#value'  => $this->t( 'Save' ),
					'#name'   => 'submit_config',
					'#submit' => array( '::submitConfig' ),
				),
			);
		}

		$form['#attached']['library'][] = 'webt/setting_tab.style';

		return $form;
	}

	/**
	 * Function executed on valid authorization setting save event. Saves submitted data in Drupal config.
	 *
	 * @param array              $form Submitted form.
	 * @param FormStateInterface $form_state Submitted form's state.
	 * @return void
	 */
	public function submitForm( array &$form, FormStateInterface $form_state ) {
		\Drupal::messenger()->deleteAll();
		$config    = \Drupal::getContainer()->get( 'config.factory' )->getEditable( 'webt.settings' );
		$mt_engine = $form_state->getValue( 'mt_engine' );
		$config->set( 'mt_engine', $mt_engine );

		if ( $this->etranslation_engine_id === $mt_engine ) {
			$config->set( 'etranslation_appname', $form_state->getValue( 'etranslation_appname' ) );
			if ( $config->get( 'etranslation_password' ) !== $form_state->getValue( 'etranslation_password' ) ) {
				$encrypted_pwd = EtranslationUtils::encrypt_password( $form_state->getValue( 'etranslation_password' ) );
				$config->set( 'etranslation_password', $encrypted_pwd );
			}
		} elseif ( $this->generic_mt_engine_id === $mt_engine ) {
			$config->set( 'mt_api_url', $form_state->getValue( 'mt_api_url' ) );
			$config->set( 'mt_api_key', $form_state->getValue( 'mt_api_key' ) );
		}
		$config->clear( 'language_directions' );

		$config->save();
		\Drupal::messenger()->addStatus( t( 'Authorization settings saved.' ) );
		$this->checkMachineTranslatonConnection();
	}

	/**
	 * Validation function executed on authorization setting save event. Currently checks if Custom provider URL is entered correctly.
	 *
	 * @param array              $form Form.
	 * @param FormStateInterface $form_state Form state.
	 * @return void
	 */
	public function validateForm( array &$form, FormStateInterface $form_state ) {
		$engine = $form_state->getValue( 'mt_engine' );
		switch ( $engine ) {
			case $this->etranslation_engine_id:
				return;
			case $this->generic_mt_engine_id:
				$url_key    = 'mt_api_url';
				$mt_api_url = $form_state->getValue( $url_key );
				if ( ! filter_var( $mt_api_url, FILTER_VALIDATE_URL ) ) {
					$form_state->setErrorByName( $url_key, $this->t( 'Please enter a valid URL' ) );
				}
				break;
			default:
				\Drupal::logger( WEBT_LOGGER )->error( "Invalid translation engine: '$engine'. Please recheck your MT provider authorization settings!" );
				return;
		}
	}

	/**
	 * Function executed on valid configuration setting save event. Saves submitted data in Drupal config.
	 *
	 * @param array              $form Submitted form.
	 * @param FormStateInterface $form_state Submitted form's state.
	 * @return void
	 */
	public function submitConfig( array &$form, FormStateInterface $form_state ) {
		\Drupal::messenger()->deleteAll();
		$config = \Drupal::getContainer()->get( 'config.factory' )->getEditable( 'webt.settings' );

		$mt_engine = $form_state->getValue( 'mt_engine' );

		if ( $this->etranslation_engine_id === $mt_engine ) {
			$config->set( 'etranslation_timeout', $form_state->getValue( 'etranslation_advanced' )['etranslation_timeout'] );
			$config->set( 'etranslation_chars_per_request', $form_state->getValue( 'etranslation_advanced' )['etranslation_chars_per_request'] );
		} elseif ( $this->generic_mt_engine_id === $mt_engine ) {
			$config->set( 'mt_chars_per_request', $form_state->getValue( 'mt_advanced' )['mt_chars_per_request'] );
		}
		$langcodes = array_keys( \Drupal::languageManager()->getLanguages() );
		foreach ( $langcodes as $langcode ) {
			$domain = $form_state->getValue( "mt_domain_$langcode" );
			$config->set( "mt_domain_$langcode", $domain );
		}
		$config->save();
		\Drupal::messenger()->addStatus( t( 'Provider configuration saved.' ) );
	}

	/**
	 * Checks whether machine translation authorization data is provided for selected MT provider.
	 *
	 * @return boolean
	 */
	private function machineTranslationConfigured() {
		$config = $this->config( 'webt.settings' );
		$engine = $config->get( 'mt_engine' );
		switch ( $engine ) {
			case $this->etranslation_engine_id:
				return $config->get( 'etranslation_appname' ) && $config->get( 'etranslation_password' );
			case $this->generic_mt_engine_id:
				return $config->get( 'mt_api_url' ) && $config->get( 'mt_api_key' );
			default:
				return false;
		}
	}

	/**
	 * Checks if connection with machine translation service is successful.
	 *
	 * @return boolean
	 */
	function checkMachineTranslatonConnection() {
		$translation_manager   = \Drupal::service( 'webt.translation_manager' );
		$translation_service   = $translation_manager->translation_service;
		$connection_successful = $translation_service->test_connection();
		if ( $connection_successful ) {
			\Drupal::messenger()->addStatus( t( 'Machine-translation connection successful' ) );
			return true;
		} else {
			$logs = \Drupal\Core\Link::fromTextAndUrl( 'logs', \Drupal\Core\Url::fromRoute( 'dblog.overview' ) );
			\Drupal::messenger()->addError( t( 'There was a problem connecting to machine-translation provider. Check your MT configuration, @logs or try again later.', array( '@logs' => $logs->toString() ) ) );
			return false;
		}
	}

	/**
	 * Creates dropdown field container for machine translation engine selection. Adds warning if language not supported by MT provider.
	 *
	 * @return array
	 */
	private function getMtEngineFields() {
		$config                = $this->config( 'webt.settings' );
		$translation_manager   = \Drupal::service( 'webt.translation_manager' );
		$translation_service   = $translation_manager->translation_service;
		$supported_languages   = $translation_service->get_supported_languages();
		$unsupported_languages = array();

		$langcodes = \Drupal::languageManager()->getLanguages();
		$selects   = array();
		foreach ( $langcodes as $langcode => $value ) {
			$directions = array_filter(
				$supported_languages,
				function ( $direction ) use ( $langcode ) {
					return $direction->trgLang === $langcode;
				}
			);
			$options    = array( null => t( '[Unspecified]' ) );
			foreach ( $directions as $direction ) {
				$options[ $direction->domain ] = isset( $direction->name ) ? $direction->name : $direction->domain;
			}
			$select                           = array(
				'#title'         => $value->getName(),
				'#type'          => 'select',
				'#options'       => $options,
				'#default_value' => $config->get( "mt_domain_$langcode" ),
			);
			$selects[ "mt_domain_$langcode" ] = $select;

			if ( 1 === count( $options ) ) {
				$unsupported_languages[] = $value->getName();
			}
		}
		if ( count( $unsupported_languages ) > 0 ) {
			$localized_languages = array_map(
				function( $l ) {
					return t( $l );
				},
				$unsupported_languages
			);
			$list_str            = implode( ', ', $localized_languages );
			\Drupal::messenger()->addWarning( t( 'Current MT provider does not support following languages:' ) . ' ' . $list_str . '!' );
		}
		return array(
			'#type'       => 'container',
			'selects'     => $selects,
			'#attributes' => array( 'class' => 'mt_engines_container' ),
		);
	}
}
