<?php

namespace Drupal\webt;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\webt\Model\ProcessingResponse;
use Drupal\webt\Model\StringType;
use Drupal\webt\Model\TranslationStatus;

/**
 * Service that creates/updates/deletes translated versions of source items (entities, ui, config strings).
 */
class TranslationManager {

	/**
	 * Machine translation service that translates strings
	 *
	 * @var Drupal\webt\translation_engines\AbstractTranslationService
	 */
	public $translation_service;

	/**
	 * Drupal config factory
	 *
	 * @var Drupal\Core\Config\ConfigFactoryInterface
	 */
	protected $config_factory;

	/**
	 * Initializes TranslationManager
	 *
	 * @param ConfigFactoryInterface $config_factory Config factory.
	 */
	public function __construct( ConfigFactoryInterface $config_factory ) {
		$this->config_factory = $config_factory;
		$engine               = \Drupal::config( 'webt.settings' )->get( 'mt_engine' );
		if ( ! $engine ) {
			$engine = 'etranslation';
		}
		$this->translation_service = \Drupal::service( 'webt.translation_service.' . $engine );
	}

	/**
	 * Create translations for given entities in given target languages.
	 *
	 * @param array    $entities Entities to be translated.
	 * @param string   $type     Entity type needed for progress bar message updates.
	 * @param string[] $target_langcodes Target language codes.
	 * @param boolean  $skip_if_translated By default functiojns skips already translated entities. Retranslates them if $skip_if_translated is false.
	 * @return integer of ProcessingResponse
	 */
	public function translate_entities( $entities, $type, $target_langcodes, $skip_if_translated = true ) {
		$translatable_values   = array();
		$translatable_enitites = array();

		$translated_versions_to_create = 0;
		$translated_versions_created   = 0;

		// group entities by src-trg language.
		foreach ( $entities as $entity ) {
			$source_langcode = $entity->language()->getId();
			foreach ( $target_langcodes as $langcode ) {
				if ( $langcode === $source_langcode ) {
					continue;
				}
				if ( $skip_if_translated && $entity->hasTranslation( $langcode ) ) {
					continue;
				}
				$key = "$source_langcode->$langcode";

				if ( ! isset( $translatable_values[ $key ] ) ) {
					$translatable_values[ $key ]   = array();
					$translatable_enitites[ $key ] = array();
				}

				$translatable_fields = $this->get_translatable_fields( $entity );
				if ( ! empty( $translatable_fields ) ) {
					$translated_versions_to_create++;
					$entity_values                   = $this->extract_translatable_values( $entity, $translatable_fields );
					$translatable_values[ $key ]     = array_merge( $translatable_values[ $key ], $entity_values );
					$translatable_enitites[ $key ][] = $entity;
				}
			}
		}

		// translate entities for each src-trg language pair.
		foreach ( $translatable_values as $key => $strings_to_translate ) {
			$key_parts            = explode( '->', $key );
			$src_lang             = $key_parts[0];
			$trg_lang             = $key_parts[1];
			$translation_response = $this->translation_service->translate( $src_lang, $trg_lang, $strings_to_translate, $type );

			if ( empty( $translation_response->translations ) || TranslationStatus::FULL_RESULT !== $translation_response->status ) {
				continue;
			}

			$offset = 0;
			foreach ( $translatable_enitites[ $key ] as $entity ) {
				$entity_translations        = array();
				$entity_translatable_fields = $this->get_translatable_fields( $entity );
				$translatable_value_count   = count( $this->extract_translatable_values( $entity, $translatable_fields ) );
				for ( $i = 0; $i < $translatable_value_count; $i++, $offset++ ) {
					$entity_translations[] = $translation_response->translations[ $offset ];
				}
				if ( ! empty( $entity_translations ) ) {
					$this->entity_create_translation( $entity, $trg_lang, $entity_translatable_fields, $entity_translations );
					$translated_versions_created++;
				}
			}
		}

		return $this->get_processing_response( $translated_versions_created, $translated_versions_to_create );
	}

	/**
	 * Creates or updates all translations of an enitity
	 *
	 * @param \Drupal\Core\Entity\Entity $entity Entity to translate.
	 * @return void
	 */
	public function entity_update_all_translations( $entity ) {
		\Drupal::logger( WEBT_LOGGER )->info( 'Updating all translations for entity' );
		$source_langcode = $entity->language()->getId();
		$langcodes       = \Drupal::languageManager()->getLanguages();
		$langcodes_list  = array_keys( $langcodes );

		foreach ( $langcodes_list as $langcode ) {
			if ( $langcode === $source_langcode || ! $this->translation_service->is_language_supported( $langcode ) ) {
				continue;
			}
			$this->translate_entity( $entity, $source_langcode, $langcode );
		}
	}

	/**
	 * Translate entity from specified source language to specified target language.
	 *
	 * @param \Drupal\Core\Entity\Entity $entity Enitity to translate.
	 * @param string                     $src_lang Source language code.
	 * @param string                     $trg_lang Target language code.
	 * @param string[]                   $translations If provided, uses given string translations and skips machine translation part.
	 * @return void
	 */
	public function translate_entity( $entity, $src_lang, $trg_lang, $translations = null ) {
		\Drupal::logger( WEBT_LOGGER )->info( "Translating entity ($src_lang-$trg_lang)" );
		$translatable_fields = $this->get_translatable_fields( $entity );

		if ( count( $translatable_fields ) > 0 && $entity->isTranslatable() ) {
			$strings_to_translate = $this->extract_translatable_values( $entity, $translatable_fields );

			if ( ! $translations ) {
				$translation_response = $this->translation_service->translate( $src_lang, $trg_lang, $strings_to_translate, 'entity', false, $entity->id() );
				$translations         = $translation_response->translations;
				if ( TranslationStatus::FULL_RESULT !== $translation_response->status ) {
					return;
				}
			}
			$this->entity_create_translation( $entity, $trg_lang, $translatable_fields, $translations );
		}
	}

	/**
	 * Delete translations of entities in specified languages.
	 *
	 * @param \Drupal\Core\Entity\Entity[] $entities Entities to delete translations of.
	 * @param string[]                     $langcodes Language codes of translations which need to be deleted.
	 * @return integer of ProcessingResponse
	 */
	public function delete_translations_of_entities( $entities, $langcodes ) {
		foreach ( $entities as $entity ) {
			$source_langcode = $entity->language()->getId();
			foreach ( $langcodes as $langcode ) {
				if ( $langcode === $source_langcode ) {
					continue;
				}
				if ( $entity->hasTranslation( $langcode ) ) {
					$entity->removeTranslation( $langcode );
					$entity->save();
				}
			}
		}
		return ProcessingResponse::$fully_processed;
	}

	/**
	 * Translate given UI strings and save translations.
	 *
	 * @param \Drupal\locale\TranslationString[] $translatable_strings Translatable UI strings.
	 * @param string                             $src_lang Source language of UI strings.
	 * @param string                             $trg_lang Target language.
	 * @return integer of ProcessingResponse
	 */
	public function translate_ui_strings( $translatable_strings, $src_lang, $trg_lang ) {
		$locale_storage = \Drupal::service( 'locale.storage' );

		$values               = array_map(
			function( $s ) {
				return $s->source;
			},
			$translatable_strings
		);
		$translation_response = $this->translation_service->translate( $src_lang, $trg_lang, $values, StringType::UI, true );
		$translations         = $translation_response->translations;
		$count                = count( $translations );

		for ( $i = 0; $i < $count; $i++ ) {
			$translated_string = $translations[ $i ];
			$translation       = $locale_storage->createTranslation(
				array(
					'lid'         => $translatable_strings[ $i ]->lid,
					'language'    => $trg_lang,
					'translation' => $translated_string,
				)
			);
			$translation->save();
		}
		return $this->get_processing_response( $count, count( $values ) );
	}

	/**
	 * Get list of all untranslated UI strings for a language.
	 *
	 * @param string $langcode Translation language code.
	 * @return \Drupal\locale\TranslationString[]
	 */
	public function get_untranslated_ui_strings( $langcode ) {
		$locale_storage        = \Drupal::service( 'locale.storage' );
		$existing_translations = $locale_storage->getTranslations( array( 'language' => $langcode ) );

		$untranslated_strings = array_filter(
			$existing_translations,
			function( $s ) {
				$has_source      = isset( $s->source ) && ! empty( $s->source );
				$has_translation = isset( $s->translation );
				return $has_source && ! $has_translation;
			}
		);
		return array_values( $untranslated_strings );
	}

	/**
	 * Delete all UI translations for given languages.
	 *
	 * @param string[] $langcodes Language codes to delete translations of.
	 * @return integer of ProcessingResponse
	 */
	public function delete_ui_translations( $langcodes ) {
		$source_langcode = $this->get_default_language_code();
		$storage         = \Drupal::service( 'locale.storage' );

		foreach ( $langcodes as $langcode ) {
			if ( $langcode === $source_langcode ) {
				continue;
			}
			$storage->deleteTranslations( array( 'language' => $langcode ) );
		}
		return ProcessingResponse::$fully_processed;
	}

	/**
	 * Translate given configuration strings to a language.
	 *
	 * @param array  $config_strings Translatable configuration string items.
	 * @param string $langcode Language code to translate to.
	 * @return integer of ProcessingResponse
	 */
	public function translate_config_strings( $config_strings, $langcode ) {
		// prepare source value list for translation.
		$source_strings = array();
		foreach ( $config_strings as $item ) {
			array_walk_recursive(
				$item,
				function( $p ) use ( &$source_strings ) {
					if ( is_string( $p ) ) {
						$source_strings[] = $p;
					}
				}
			);
		}
		$offset = 0;
		if ( ! empty( $source_strings ) ) {
			// translate.
			$source_langcode      = $this->get_default_language_code();
			$translation_response = $this->translation_service->translate( $source_langcode, $langcode, $source_strings, StringType::CONFIG, true );
			$translations         = $translation_response->translations;

			// replace source values with translations & save.
			if ( ! empty( $translations ) && TranslationStatus::FULL_RESULT === $translation_response->status ) {
				$collection = \Drupal::service( 'config.storage' )->createCollection( "language.$langcode" );
				foreach ( $config_strings as $key => $value ) {
					$translated_data = $value;
					array_walk_recursive(
						$translated_data,
						function( &$v ) use ( $translations, &$offset ) {
							if ( is_string( $v ) ) {
								$v = $translations[ $offset ];
								$offset++;
							}
						}
					);
					$collection->write( $key, $translated_data );
				}
			}
		}

		return $this->get_processing_response( $offset, count( $source_strings ) );
	}

	/**
	 * Gets all untranslated cofiguration strings for a language.
	 *
	 * @param string $langcode Language code.
	 * @return array
	 */
	public function get_untranslated_config_strings( $langcode ) {
		$translatable_items    = $this->get_translatable_config_strings();
		$translated_item_names = \Drupal::service( 'config.storage' )->createCollection( "language.$langcode" )->listAll();

		// get untranslated config item names.
		$missing_items = array_filter(
			$translatable_items,
			function( $name ) use ( $translated_item_names ) {
				return ! in_array( $name, $translated_item_names, true );
			},
			ARRAY_FILTER_USE_KEY
		);

		return $missing_items;
	}

	/**
	 * Get all translatable configuration strings (in source language).
	 *
	 * @return array
	 */
	private function get_translatable_config_strings() {
		$source_item_names  = $this->config_factory->listAll();
		$config_items       = $this->config_factory->loadMultiple( $source_item_names );
		$translatable_items = array();

		// identify translatable configuration items and their fields.
		foreach ( $config_items as $item ) {
			$original = $item->getOriginal();
			if ( isset( $original['langcode'] ) ) {
				$translatable_fields = $this->get_translatable_config_fields( $original );
				if ( ! empty( $translatable_fields ) ) {
					$translatable_items[ $item->getName() ] = $translatable_fields;
				}
			}
		}

		return $translatable_items;
	}

	/**
	 * Get all translatable properties & their values for a configuration string item.
	 *
	 * @param mixed $config_item Configuration string item.
	 * @return array
	 */
	private function get_translatable_config_fields( $config_item ) {
		$translatable_fields = array();
		// keys that usually hold translatable values.
		$search_keys = array( 'label', 'name', 'description', 'message' );
		// arrays that can contain $search_keys.
		$container_keys = array( 'settings' );

		foreach ( $search_keys as $key ) {
			// check 1st level values.
			if ( isset( $config_item[ $key ] ) && strlen( $config_item[ $key ] ) > 0 ) {
				if ( 'name' === $key && in_array( 'label', array_keys( $translatable_fields ), true ) ) {
					// do not translate 'name' if 'label' already present. Workaround for drupal/admin/config/media/image-styles translation.
					continue;
				}
				$translatable_fields[ $key ] = $config_item[ $key ];
			}
			// check 2nd level values.
			foreach ( $container_keys as $container_key ) {
				if ( isset( $config_item[ $container_key ] ) && isset( $config_item[ $container_key ][ $key ] ) &&
					strlen( $config_item[ $container_key ][ $key ] ) > 0 ) {
					if ( ! isset( $translatable_fields[ $container_key ] ) ) {
						$translatable_fields[ $container_key ] = array();
					}
					$translatable_fields[ $container_key ][ $key ] = $config_item[ $container_key ][ $key ];
				}
			}
		}

		return $translatable_fields;
	}

	/**
	 * Delete all configuration string translations for given languages.
	 *
	 * @param string[] $langcodes Target language codes.
	 * @return integer of ProcessingResponse
	 */
	public function delete_config_translations( $langcodes ) {
		$source_langcode = $this->get_default_language_code();

		foreach ( $langcodes as $langcode ) {
			if ( $langcode === $source_langcode ) {
				continue;
			}
			\Drupal::service( 'config.storage' )->createCollection( "language.$langcode" )->deleteAll();
		}
		return ProcessingResponse::$fully_processed;
	}

	/**
	 * Get translatable fields for given entity.
	 *
	 * @param \Drupal\Core\Entity\Entity $entity Entity.
	 * @return array
	 */
	public function get_translatable_fields( $entity ) {
		switch ( $entity->getEntityTypeId() ) {
			case 'node':
				return array(
					'title'       => array( 'value' ),
					'body'        => array( 'value', 'summary' ),
					'field_image' => array( 'alt' ),
				);
			case 'comment':
				return array(
					'subject'      => array( 'value' ),
					'comment_body' => array( 'value' ),
				);
			case 'taxonomy_term':
				return array( 'name' => array( 'value' ) );
			default:
				// type not implemented.
				return array();
		}
	}

	/**
	 * Checks whether translatable content of entity has changed and it needs retranslation.
	 *
	 * @param \Drupal\Core\Entity\Entity $entity Entity.
	 * @return boolean
	 */
	public function entity_translatable_content_changed( $entity ) {
		$translatable_fields = $this->get_translatable_fields( $entity );

		if ( count( $translatable_fields ) > 0 ) {
			foreach ( $translatable_fields as $parent => $fields ) {
				foreach ( $fields as $field ) {
					if ( ! $entity->original || ( $entity->$parent && $entity->$parent->$field !== $entity->original->$parent->$field ) ) {
						return true;
					}
				}
			}
		}

		if ( $entity->hasField( 'field_tags' ) && $entity->field_tags->referencedEntities() !== $entity->original->field_tags->referencedEntities() ) {
			return true;
		}
		if ( $entity->hasField( 'field_image' ) && $entity->field_image->entity !== $entity->original->field_image->entity ) {
			return true;
		}
		return false;
	}

	/**
	 * Calculate translation progress percentage for given content type and language.
	 *
	 * @param string $type Content type.
	 * @param string $langcode Target language code.
	 * @return float
	 */
	public function get_translation_progress_percents( $type, $langcode ) {
		$source_language  = $this->get_default_language_code();
		$source_count     = 0;
		$translated_count = 0;
		switch ( $type ) {
			case StringType::UI:
				if ( $source_language === $langcode ) {
					return 100;
				}
				$locale_storage   = \Drupal::service( 'locale.storage' );
				$source_count     = count( $locale_storage->getStrings() );
				$translated_count = $source_count - count( $this->get_untranslated_ui_strings( $langcode ) );
				break;
			case StringType::CONFIG:
				if ( $source_language === $langcode ) {
					return 100;
				} else {
					$source_strings   = $this->get_translatable_config_strings();
					$source_count     = count( $source_strings );
					$translated_count = $source_count - count( $this->get_untranslated_config_strings( $langcode ) );
					break;
				}
			default:
				$source_strings     = \Drupal::entityTypeManager()
					->getStorage( $type )
					->loadMultiple();
				$translated_strings = array_filter(
					$source_strings,
					function( $e ) use ( $langcode ) {
						return $e->language()->getId() === $langcode || $e->hasTranslation( $langcode );
					}
				);
				$source_count       = count( $source_strings );
				$translated_count   = count( $translated_strings );
				break;
		}
		return $source_count > 0 ? (float) 100 * $translated_count / $source_count : 0;
	}

	/**
	 * Get default language code of the website.
	 *
	 * @return string
	 */
	public function get_default_language_code() {
		return \Drupal::service( 'language.default' )->get()->getId();
	}

	/**
	 * Create translated version of an entity.
	 *
	 * @param \Drupal\Core\Entity\Entity $entity Entity to translate.
	 * @param string                     $trg_lang Target language code.
	 * @param array                      $translation_fields Translatable fields for this entity.
	 * @param array                      $translation_values Translated values of translatable fields.
	 * @return void
	 */
	private function entity_create_translation( $entity, $trg_lang, $translation_fields, $translation_values ) {
		$translation_count  = count( $translation_values );
		$source_value_count = count( $this->extract_translatable_values( $entity, $translation_fields ) );
		if ( count( $translation_fields ) > 0 && $source_value_count === $translation_count && $entity->isTranslatable() ) {
			if ( $entity->hasTranslation( $trg_lang ) ) {
				$entity->removeTranslation( $trg_lang );
			}

			$field_translations = array();
			$offset             = 0;
			foreach ( $translation_fields as $parent => $fields ) {
				$field_translations[ $parent ] = array();
				foreach ( $fields as $field ) {
					$field_translations[ $parent ][ $field ] = $translation_values[ $offset ];
					$offset++;
				}
			}

			$translation = $entity->addTranslation( $trg_lang, $field_translations );

			$body_key = null;
			switch ( $entity->getEntityTypeId() ) {
				case 'node':
					$body_key = 'body';
					break;
				case 'comment':
					$body_key = 'comment_body';
					break;
			}

			if ( $body_key ) {
				$translation->$body_key->format = $entity->$body_key->format;
			}

			if ( $entity->hasField( 'field_tags' ) ) {
				$translation->field_tags = $entity->field_tags;
			}

			if ( $entity->hasField( 'field_image' ) ) {
				$translation->set(
					'field_image',
					array(
						'target_id' => $entity->field_image->target_id,
						'alt'       => $translation->field_image->alt,
					)
				);
			}

			if ( $entity->hasField( 'status' ) ) {
				$translation->status->value = $entity->status->value;
			}

			if ( $entity->hasField( 'content_translation_source' ) ) {
				$translation->content_translation_source->value = $entity->langcode->value;
			}

			$translation->save();
		}
	}

	/**
	 * Calculate processing response status based on translation and reference count.
	 *
	 * @param integer $translation_count Translated string value count.
	 * @param integer $reference_value_count Original (reference) string value count.
	 * @return integer of ProcessingResponse
	 */
	private function get_processing_response( $translation_count, $reference_value_count ) {
		if ( 0 === $reference_value_count ) {
			return ProcessingResponse::$already_processed;
		} elseif ( $translation_count === $reference_value_count ) {
			return ProcessingResponse::$fully_processed;
		} elseif ( 0 === $translation_count && 0 < $reference_value_count ) {
			return ProcessingResponse::$processing_error;
		} else {
			return ProcessingResponse::$partially_processed;
		}
	}

	/**
	 * Extract translatable values from an entity using list of translatable fields.
	 *
	 * @param \Drupal\Core\Entity\Entity $entity Entity.
	 * @param array                      $translatable_fields Translatable fields of given entity.
	 * @return string[]
	 */
	public function extract_translatable_values( $entity, $translatable_fields ) {
		$translatable_values = array();
		foreach ( $translatable_fields as $parent => $fields ) {
			foreach ( $fields as $field ) {
				$translatable_values[] = $entity->$parent->$field ?? '';
			}
		}
		return $translatable_values;
	}
}
