<?php

namespace Drupal\webt\translation_engines;

use Drupal\Core\Http\ClientFactory;
use Drupal\Component\Utility\Html;
use Drupal\webt\Controller\PretranslationProgressController;
use Drupal\webt\Model\TranslationResponse;
use Drupal\webt\Model\TranslationStatus;

/**
 * Translation service base class.
 */
abstract class AbstractTranslationService {

	/**
	 * Array of regex patterns for non-translatable Drupal variables in UI strings.
	 *
	 * @var array
	 */
	private $drupal_variable_patterns = array(
		// match @.. and %.. and !..
		'(([^(@|%|!)^\sa-zA-Z\d+<>]*)(@|%|!)([^\s<>(@|%|!)]+))',
		// match curly bracket variables like {var}.
		'({([^\s]*?)})',
		// match double curly bracket variables like {{ var }}.
		'({{\s*(.*?)\s*}})',
		// match square bracket variables like [site:name].
		'(\[(.*?)\:(.*?)\])',
		// match PHP variables like $var['abc'].
		'(\$(([a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*(->)*)*(\[[^\]]*\])*))',
	);

	/**
	 * HTML wrapper prefix for every translatable string
	 *
	 * @var string
	 */
	private $html_wrapper_prefix = '<div>';
	/**
	 * HTML wrapper suffix for every translatable string
	 *
	 * @var string
	 */
	private $html_wrapper_suffix = '</div>';
	/**
	 * Special character dictionary for chars to replace before string is converted to HTML.
	 *
	 * @var array
	 */
	private $special_char_dict = array( "\x03" => '<etx></etx>' );
	/**
	 * List of translatable HTML attributes.
	 *
	 * @var array
	 */
	private $translatable_html_attributes = array( 'alt' );

	/**
	 * HTTP client
	 *
	 * @var \GuzzleHttp\Client
	 */
	protected $http_client;
	/**
	 * Whether to retry entity translation on error
	 *
	 * @var boolean
	 */
	protected $retry_on_entity_translation_error = true;
	/**
	 * Char limit which determines when to split translatable strings into separate request
	 *
	 * @var integer
	 */
	protected $max_chars_per_request = 60000;
	/**
	 * Drupal configuration
	 *
	 * @var \Drupal\Core\Config\ImmutableConfig
	 */
	protected $config;

	/**
	 * Initializes service.
	 *
	 * @param ClientFactory $http_client_factory HTTP client factory.
	 */
	public function __construct( ClientFactory $http_client_factory ) {
		$this->http_client = $http_client_factory->fromOptions( array( 'timeout' => 120 ) );
		$this->config      = \Drupal::config( 'webt.settings' );
	}

	/**
	 * Translation request function implemented by each MT provider.
	 *
	 * @param string   $from       Source langcode.
	 * @param string   $to         Target langcode.
	 * @param string[] $values     Translatable XML strings.
	 * @param string   $entity_id  Related entity ID.
	 * @return string[]
	 */
	abstract protected function send_translation_request( $from, $to, $values, $entity_id );

	/**
	 * Language direction request function implemented by each MT provider.
	 *
	 * @return Psr\Http\Message\ResponseInterface
	 */
	abstract protected function send_language_direction_request();

	/**
	 * Checks if MT authorization data is provided.
	 *
	 * @return boolean
	 */
	abstract public function is_authorized();

	/**
	 * Retrieves languages supported by MT provider either from cache or by making an HTTP request.
	 *
	 * @return array
	 */
	public function get_supported_languages() {
		$key         = 'language_directions';
		$config      = \Drupal::getContainer()->get( 'config.factory' )->getEditable( 'webt.settings' );
		$json_string = $config->get( $key );
		if ( ! $json_string ) {
			$response = $this->send_language_direction_request();
			if ( $response && 200 === $response->getStatusCode() ) {
				$json_string = $this->map_language_direction_response_body( $response->getBody() );
				$config->set( $key, $json_string );
				$config->save();
			} else {
				$result = array();
				$config->set( $key, $result );
				$config->save();
				return $result;
			}
		}
		$obj = json_decode( $json_string );
		return $obj->languageDirections ?? array();
	}

	/**
	 * Checks if given langcode is supported by MT provider.
	 *
	 * @param string $langcode Language code.
	 * @return boolean
	 */
	public function is_language_supported( $langcode ) {
		$lang                = substr( $langcode, 0, 2 );
		$supported_languages = $this->get_supported_languages();
		foreach ( $supported_languages as $engine ) {
			if ( $engine->trgLang === $lang ) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Translates given string values
	 *
	 * @param string   $from Source langcode.
	 * @param string   $to Target langcode.
	 * @param string[] $values String array of translatable values.
	 * @param string   $object_type String content type related to given translatable values.
	 * @param boolean  $encode_placeholders Whether to encode non-translatable Drupal variables before translation.
	 * @param string   $entity_id Related entity ID.
	 * @return string[]
	 */
	public function translate( $from, $to, $values, $object_type = '', $encode_placeholders = false, $entity_id = null ) {
		$current_char_len    = 0;
		$request_value_lists = array();
		$current_list        = array();
		$lang_from           = substr( $from, 0, 2 );
		$lang_to             = substr( $to, 0, 2 );

		// split into multiple arrays not exceeding max char limit.
		foreach ( $values as $value ) {
			$chars = strlen( $value );
			if ( $current_char_len + $chars > $this->max_chars_per_request ) {
				$request_value_lists[] = $current_list;
				$current_list          = array( $value );
				$current_char_len      = $chars;
			} else {
				$current_list[]    = $value;
				$current_char_len += $chars;
			}
		}
		if ( ! empty( $current_list ) ) {
			$request_value_lists[] = $current_list;
		}

		// send requests and merge results together.
		$full_result   = array();
		$request_count = count( $request_value_lists );

		$percentage_completed = 0;
		$reference_id         = $this->update_progress_info( 0, $object_type, $lang_from, $lang_to, $entity_id );

		for ( $i = 0; $i < $request_count; $i++ ) {
			$this->check_client_disconnected( $reference_id, $entity_id );
			// replace necessary special characters removed on HTML tree creation; currently needed only for UI strings containing multiple values in same string, separated by ETX char.
			$request_value_lists[ $i ] = str_replace( array_keys( $this->special_char_dict ), $this->special_char_dict, $request_value_lists[ $i ] );

			$contains_only_empty_strings = empty(
				array_filter(
					$request_value_lists[ $i ],
					function ( $a ) {
						return ! empty( $a );
					}
				)
			);

			if ( $contains_only_empty_strings ) {
				$full_result = array_merge( $full_result, $request_value_lists[ $i ] );
			} else {
				// convert string[] to XML[].
				$xml_values = $this->translatable_values_to_xml( $request_value_lists[ $i ] );

				$translatable_values = $xml_values;
				$encode_response     = null;
				if ( $encode_placeholders ) {
					// replace non-translatable string placeholders with XML tags (for UI strings).
					$encode_response     = $this->encode_placeholders( $xml_values );
					$translatable_values = $encode_response->values;
				}

				// translate XML[].
				\Drupal::logger( WEBT_LOGGER )->debug( "Sending $lang_from-$lang_to translation request " . ((string) ($i + 1)) . '/' . (string) $request_count . ' (' . ((string) count( $translatable_values )) . ' strings)' );
				$xml_translations = $this->send_translation_request( $lang_from, $lang_to, $translatable_values, $entity_id );

				if ( ! $xml_translations || empty( $xml_translations ) ) {
					$retries_left = 2;
					while ( $retries_left > 0 && ( ! $entity_id || $this->retry_on_entity_translation_error ) ) {
						$this->check_client_disconnected( $reference_id, $entity_id );
						\Drupal::logger( WEBT_LOGGER )->debug( "Translation request failed, retrying... (retries left: $retries_left)" );
						$xml_translations = $this->send_translation_request( $lang_from, $lang_to, $translatable_values, $entity_id );
						$retries_left--;
					}

					// do not continue if one of requests fail.
					if ( ! $xml_translations || empty( $xml_translations ) ) {
						if ( $request_count > 1 ) {
							\Drupal::logger( WEBT_LOGGER )->error( 'One of translation requests failed. Please check if machine translation provider API is working or reduce the "Max characters per single request" value (in Translation provider configuration) and try again!' );
						}
						return new TranslationResponse( $full_result, empty( $full_result ) ? TranslationStatus::NO_RESULT : TranslationStatus::PARTIAL_RESULT );
					}
				}

				if ( $encode_response ) {
					// restore non-translatable string placeholders.
					$xml_translations = $this->restore_spaces_around_placeholders( $translatable_values, $xml_translations, $encode_response->placeholders );
					$xml_translations = $this->decode_placeholders( $xml_translations, $encode_response->placeholders );
				}				
				// convert XML[] back to string[].
				$decoded_translations = $this->decode_xml_translations( $request_value_lists[ $i ], $xml_translations );
				// restore special characters (for some UI strings).
				$decoded_translations = str_replace( $this->special_char_dict, array_keys( $this->special_char_dict ), $decoded_translations );
				// merge with previous translations.
				$full_result = array_merge( $full_result, $decoded_translations );
			}
			// update percentage for pre-translation progress bar.
			$percentage_completed += 100 / $request_count;
			$this->update_progress_info( $percentage_completed, $object_type, $lang_from, $lang_to, $entity_id );
		}
		return new TranslationResponse( $full_result, TranslationStatus::FULL_RESULT );
	}

	/**
	 * Makes a language direction request to check if MT provider connection is successful.
	 *
	 * @return boolean
	 */
	public function test_connection() {
		$response = $this->send_language_direction_request();
		if ( ! $response ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'Response object is null. Please check if MT service can be reached.' );
			return false;
		}
		$content_type = isset( $response->getHeaders()['Content-Type'] ) ? $response->getHeaders()['Content-Type'][0] : null;
		if ( $content_type && ! str_contains( $content_type, 'application/json' ) ) {
			\Drupal::logger( WEBT_LOGGER )->error( "Invalid content type '$content_type' in language direction response!" );
			return false;
		}
		return 200 === $response->getStatusCode();
	}

	/**
	 * Maps language direction response to Generic API JSON response.
	 *
	 * @param string $body Response body.
	 * @return string
	 */
	protected function map_language_direction_response_body( $body ) {
		return (string) $body;
	}

	/**
	 * Converts translatable strings to XML strings.
	 *
	 * @param string $values Translatable string values.
	 * @return string[]
	 */
	protected function translatable_values_to_xml( $values ) {
		$xmls  = array();
		$count = 1;
		foreach ( $values as $v ) {
			$html_string = $this->wrap_as_html( $v );
			$dom         = Html::load( $html_string );

			$body  = $dom->getElementsByTagName( 'body' )[0];
			$nodes = array( $body->firstChild );

			while ( $node = array_shift( $nodes ) ) {
				if ( $node instanceof \DOMElement ) {
					// replace each non-text node with <g> node (or x tag if self-closing), also removing all attributes.
					$newTag = $dom->createElement( $node->hasChildNodes() ? 'g' : 'x' );
					// set id for each <g> or <x> node.
					$newTag->setAttribute( 'id', $count++ );
					// replace node.
					$node->parentNode->replaceChild( $newTag, $node );
					// move all children nodes to new node.
					while ( $node->hasChildNodes() ) {
						$newTag->appendChild( $node->firstChild );
					}
					// process children nodes next.
					$nodes = array_merge( iterator_to_array( $newTag->childNodes ), $nodes );

					foreach ( $this->translatable_html_attributes as $attr ) {
						if ( $node->hasAttribute( $attr ) ) {
							// MT engines may not translate XML attributes, so create temporary text nodes from attribute values.
							$a_tag = $dom->createElement( 'a' );
							$a_tag->setAttribute( 'id', "$attr-$count" );
							$a_tag->setAttribute( 'name', $attr );
							$a_tag->nodeValue = $node->getAttribute( $attr );
							$newTag->appendChild( $a_tag );
						}
					}
				}
			}
			$xmls[] = $dom->saveXML( $body->firstChild );
		}
		return $xmls;
	}

	/**
	 * Converts translated XML strings back to normal strings.
	 *
	 * @param string $source_values Original (normal) strings that were translated.
	 * @param string $translated_values XML translations.
	 * @return string[]
	 */
	public function decode_xml_translations( $source_values, $translated_values ) {
		$translation_count    = count( $translated_values );
		$decoded_translations = array();
		for ( $i = 0; $i < $translation_count; $i++ ) {
			// retrieve source and translated values.
			$source      = $source_values[ $i ];
			$translation = $translated_values[ $i ];

			// load source as html document.
			$html_source  = $this->wrap_as_html( $source );
			$html_dom     = Html::load( $html_source );
			$html_wrapper = $html_dom->getElementsByTagName( 'body' )[0]->firstChild;

			// load translation as xml document.
			$xml         = simplexml_load_string( $translation );
			$xml_dom     = dom_import_simplexml( $xml )->ownerDocument;
			$xml_wrapper = $xml_dom->firstChild;

			// move translated attribute values from nodes back to attributes.
			$this->process_attribute_nodes( $xml_dom );
			// replace text nodes in source html with matching text nodes from translation xml.
			$new_node = $this->replace_text_nodes_bottom_up( $html_wrapper, $xml_wrapper );
			// save as XML instead of HTML to avoid url encoding.
			$text_xml = $html_dom->saveXML( $new_node );
			// remove XML tag.
			$text_value = preg_replace( '/\<\?xml(.*?)\?\>/', '', $text_xml );

			// remove wrapper.
			$html_prefix_len        = strlen( $this->html_wrapper_prefix );
			$decoded_translation    = substr( $text_value, $html_prefix_len, strlen( $text_value ) - $html_prefix_len - strlen( $this->html_wrapper_suffix ) );
			$decoded_translations[] = $decoded_translation;
		}
		return $decoded_translations;
	}

	/**
	 * Remove XML attribute nodes and restore their values as parent attributes.
	 *
	 * @param \DOMDocument $dom Document.
	 * @return void
	 */
	private function process_attribute_nodes( $dom ) {
		// set parent node attributes from <a> tags containing attribute translations.
		$attribute_nodes = $dom->getElementsByTagName( 'a' );
		$count           = count( $attribute_nodes );
		for ( $i = 0; $i < $count; $i++ ) {
			$node      = $attribute_nodes[0]; // $attribute_nodes list is updated on removeChild, next element's index is always 0.
			$parent    = $node->parentNode;
			$attr_name = $node->getAttribute( 'name' );
			$attr_val  = $node->textContent;
			$parent->setAttribute( $attr_name, $attr_val );
			$parent->removeChild( $node );
		}
	}

	/**
	 * Traverse original and translated XML strings bottom up & replace original string values with their translations.
	 *
	 * @param DOMNode $original_node Original XML node.
	 * @param DOMNode $encoded_node Translated XML node.
	 * @return DOMNode
	 */
	private function replace_text_nodes_bottom_up( $original_node, $encoded_node ) {
		// Recursively call this function on each child node.
		$original_child_node = $original_node->lastChild;
		$encoded_child_node  = $encoded_node->lastChild;
		while ( $original_child_node && $encoded_child_node ) {
			// Skip any extra text nodes in original HTML (fix for Drupal 8 which inserts newlines between paragraph nodes).
			while ( $original_child_node && $encoded_child_node && XML_TEXT_NODE === $original_child_node->nodeType && XML_TEXT_NODE !== $encoded_child_node->nodeType ) {
				$original_child_node = $original_child_node->previousSibling;
			}

			$this->replace_text_nodes_bottom_up( $original_child_node, $encoded_child_node );
			$original_child_node = $original_child_node->previousSibling;
			$encoded_child_node  = $encoded_child_node->previousSibling;

			// workaround for extra spaces inserted by MT between <g> tags like: '</g> <g>'.
			while ( $original_child_node && $encoded_child_node && XML_TEXT_NODE === $encoded_child_node->nodeType && empty( trim( $encoded_child_node->nodeValue ) ) && XML_TEXT_NODE !== $original_child_node->nodeType ) {
				$textNode = $original_node->ownerDocument->createTextNode( $encoded_child_node->nodeValue );
				$original_node->insertBefore( $textNode, $original_child_node->nextSibling );
				$encoded_child_node = $encoded_child_node->previousSibling;
			}
		}

		// Process the current node.
		if ( XML_TEXT_NODE === $original_node->nodeType ) {
			$original_node->nodeValue = $encoded_node->nodeValue;
		} elseif ( $encoded_node instanceof \DOMElement ) {
			// Copy translated attributes.
			foreach ( $this->translatable_html_attributes as $attr ) {
				if ( $encoded_node->hasAttribute( $attr ) ) {
					$original_node->setAttribute( $attr, $encoded_node->getAttribute( $attr ) );
				}
			}
		}
		return $original_node;
	}

	/**
	 * Match Drupal variables with regex and encode them as XML tags.
	 *
	 * @param string[] $values XML translatable values.
	 * @return \stdClass
	 */
	private function encode_placeholders( $values ) {
		$regex                  = '/' . implode( '|', $this->drupal_variable_patterns ) . '/';
		$response               = new \stdClass();
		$response->placeholders = array();
		$response->values       = array();
		$index                  = 1;
		foreach ( $values as $v ) {
			$response->values[] = preg_replace_callback(
				$regex,
				function( $matches ) use ( &$index, $response ) {
					$key                            = $this->get_placeholder( $index++ );
					$response->placeholders[ $key ] = $matches[0];
					return $key;
				},
				$v
			);
		}
		return $response;
	}

	/**
	 * Replaces encoded placeholders with original values.
	 *
	 * @param string[] $translated_values Translated XML strings.
	 * @param string[] $placeholders Placeholder dictionary.
	 * @return string[]
	 */
	private function decode_placeholders( $translated_values, $placeholders ) {
		$translated_values = str_replace( array_keys( $placeholders ), $placeholders, $translated_values );
		return $translated_values;
	}

	/**
	 * Restores any necessary spaces around encoded placeholders removed by MT service.
	 *
	 * @param string[] $original_xml_values Original XML values.
	 * @param string[] $translated_xml_values Translated XML values.
	 * @param string[] $placeholders Placeholder dictionary.
	 * @return string[]
	 */
	private function restore_spaces_around_placeholders( $original_xml_values, $translated_xml_values, $placeholders ) {
		$delimiter      = '__WEBT__;';
		$original_xml   = implode( $delimiter, $original_xml_values );
		$translated_xml = implode( $delimiter, $translated_xml_values );

		foreach ( array_keys( $placeholders ) as $key ) {
			$original_pos    = strpos( $original_xml, $key );
			$translation_pos = strpos( $translated_xml, $key );
			if ( $original_pos > -1 && $translation_pos > -1 ) {
				$original_char_before    = substr( $original_xml, $original_pos - 1, 1 );
				$translation_char_before = substr( $translated_xml, $translation_pos - 1, 1 );
				if ( ctype_space( $original_char_before ) && ! ctype_space( $translation_char_before ) ) {
					$translated_xml = str_replace( $key, $original_char_before . $key, $translated_xml );
				}
				$original_char_after    = substr( $original_xml, $original_pos + strlen( $key ), 1 );
				$translation_char_after = substr( $translated_xml, $translation_pos + strlen( $key ), 1 );
				if ( ctype_space( $original_char_after ) && ! ctype_space( $translation_char_after ) ) {
					$translated_xml = str_replace( $key, $key . $original_char_after, $translated_xml );
				}
			}
		}
		return explode( $delimiter, $translated_xml );
	}

	/**
	 * Generates an XML tag to encode a Drupal variable (placeholder).
	 *
	 * @param integer $index Index of the placeholder.
	 * @return string
	 */
	protected function get_placeholder( $index ) {
		return "<x id=\"#$index\"/>";
	}

	/**
	 * Wrap translatable string in HTML tags.
	 *
	 * @param string $string Source string
	 * @return string
	 */
	private function wrap_as_html( $string ) {
		return $this->html_wrapper_prefix . $string . $this->html_wrapper_suffix;
	}

	/**
	 * Update pre-translation progress bar info and return pre-translation request ID.
	 *
	 * @param float  $percentage Percentage completed.
	 * @param string $type String content type.
	 * @param string $from Source language.
	 * @param string $to Target language.
	 * @param string $entity_id Related entity ID.
	 * @return string
	 */
	private function update_progress_info( $percentage, $type, $from, $to, $entity_id ) {
		if ( ! $entity_id && PretranslationProgressController::getValue() ) {
			PretranslationProgressController::updateTranslationMessage( "Translating strings of type '$type' ($from -> $to)" );
			PretranslationProgressController::updateTranslationPercentage( $percentage );
			return PretranslationProgressController::getValue()->data->request_id;
		}
		return null;
	}

	/**
	 * Check if client has disconnected and pre-translation process needs to be aborted.
	 *
	 * @param string $reference_id Pre-translation request ID.
	 * @param string $entity_id Related entity ID.
	 * @return void
	 */
	private function check_client_disconnected( $reference_id, $entity_id ) {
		$value = PretranslationProgressController::getValue();
		if ( ! $entity_id && $value && $value->data ) {
			if ( $reference_id && $reference_id !== $value->data->request_id ) {
				\Drupal::logger( WEBT_LOGGER )->debug( 'Stopping execution, parallel translation detected!' );
				die();
			}
			if ( $value->data->aborted ) {
				\Drupal::logger( WEBT_LOGGER )->debug( 'Stopping execution, client disconnected!' );
				die();
			}
		}
	}
}
