<?php

namespace Drupal\webt\translation_engines\etranslation;

use Drupal\webt\translation_engines\AbstractTranslationService;
use Drupal\Core\Http\ClientFactory;

/**
 * Machine translation service implementation for eTranslation provider.
 */
class EtranslationService extends AbstractTranslationService {

	/**
	 * Char limit which determines when to split translatable strings into separate request
	 *
	 * @var integer
	 */
	public static $default_chars_per_request = 60000;

	/**
	 * Error response code dictionary.
	 *
	 * @var array
	 */
	private static $error_map = array(
		-20000 => 'Source language not specified',
		-20001 => 'Invalid source language',
		-20002 => 'Target language(s) not specified',
		-20003 => 'Invalid target language(s)',
		-20004 => 'DEPRECATED',
		-20005 => 'Caller information not specified',
		-20006 => 'Missing application name',
		-20007 => 'Application not authorized to access the service',
		-20008 => 'Bad format for ftp address',
		-20009 => 'Bad format for sftp address',
		-20010 => 'Bad format for http address',
		-20011 => 'Bad format for email address',
		-20012 => 'Translation request must be text type, document path type or document base64 type and not several at a time',
		-20013 => 'Language pair not supported by the domain',
		-20014 => 'Username parameter not specified',
		-20015 => 'Extension invalid compared to the MIME type',
		-20016 => 'DEPRECATED',
		-20017 => 'Username parameter too long',
		-20018 => 'Invalid output format',
		-20019 => 'Institution parameter too long',
		-20020 => 'Department number too long',
		-20021 => 'Text to translate too long',
		-20022 => 'Too many FTP destinations',
		-20023 => 'Too many SFTP destinations',
		-20024 => 'Too many HTTP destinations',
		-20025 => 'Missing destination',
		-20026 => 'Bad requester callback protocol',
		-20027 => 'Bad error callback protocol',
		-20028 => 'Concurrency quota exceeded',
		-20029 => 'Document format not supported',
		-20030 => 'Text to translate is empty',
		-20031 => 'Missing text or document to translate',
		-20032 => 'Email address too long',
		-20033 => 'Cannot read stream',
		-20034 => 'Output format not supported',
		-20035 => 'Email destination tag is missing or empty',
		-20036 => 'HTTP destination tag is missing or empty',
		-20037 => 'FTP destination tag is missing or empty',
		-20038 => 'SFTP destination tag is missing or empty',
		-20039 => 'Document to translate tag is missing or empty',
		-20040 => 'Format tag is missing or empty',
		-20041 => 'The content is missing or empty',
		-20042 => 'Source language defined in TMX file differs from request',
		-20043 => 'Source language defined in XLIFF file differs from request',
		-20044 => 'Output format is not available when quality estimate is requested. It should be blank or \'xslx\'',
		-20045 => 'Quality estimate is not available for text snippet',
		-20046 => 'Document too big (>20Mb)',
		-20047 => 'Quality estimation not available',
		-40010 => 'Too many segments to translate',
		-80004 => 'Cannot store notification file at specified FTP address',
		-80005 => 'Cannot store notification file at specified SFTP address',
		-80006 => 'Cannot store translated file at specified FTP address',
		-80007 => 'Cannot store translated file at specified SFTP address',
		-90000 => 'Cannot connect to FTP',
		-90001 => 'Cannot retrieve file at specified FTP address',
		-90002 => 'File not found at specified address on FTP',
		-90007 => 'Malformed FTP address',
		-90012 => 'Cannot retrieve file content on SFTP',
		-90013 => 'Cannot connect to SFTP',
		-90014 => 'Cannot store file at specified FTP address',
		-90015 => 'Cannot retrieve file content on SFTP',
		-90016 => 'Cannot retrieve file at specified SFTP address',
	);

	/**
	 * Service API URL.
	 *
	 * @var string
	 */
	private static $api_url = 'https://webgate.ec.europa.eu/etranslation/si';

	/**
	 *  Do not retry if response has not been received within specified $etranslation_timeout on entity insert/update.
	 *
	 * @var boolean
	 */
	protected $retry_on_entity_translation_error = false;

	/**
	 * HTTP client request timeout.
	 *
	 * @var integer
	 */
	private $request_timeout = 600;
	/**
	 * Async translation response wait timeout.
	 *
	 * @var integer
	 */
	private $etranslation_timeout;

	/**
	 * XML tag which precedes every string translation
	 *
	 * @var string
	 */
	private $segment_opening = '<segment>';
	/**
	 * XML tag which closes every string translation
	 *
	 * @var string
	 */
	private $segment_closing = '</segment>';
	/**
	 * XML tag part between translations.
	 *
	 * @var string
	 */
	private $delimiter;

	/**
	 * Set service parameters from config and initialize service.
	 *
	 * @param ClientFactory $http_client_factory HTTP client factory.
	 */
	public function __construct( ClientFactory $http_client_factory ) {
		parent::__construct( $http_client_factory );
		$this->http_client           = $http_client_factory->fromOptions( array( 'timeout' => $this->request_timeout, 'verify' => false ) );
		$this->etranslation_timeout  = $this->config->get( 'etranslation_timeout' ) ? $this->config->get( 'etranslation_timeout' ) : EtranslationUtils::$etranslation_timeout_default;
		$this->max_chars_per_request = $this->config->get( 'etranslation_chars_per_request' ) ? $this->config->get( 'etranslation_chars_per_request' ) : self::$default_chars_per_request;
		$this->delimiter             = $this->segment_closing . $this->segment_opening;
	}

	/**
	 * Translate given XML strings via eTranslation
	 *
	 * @param string   $from       Source langcode.
	 * @param string   $to         Target langcode.
	 * @param string[] $xml_values Translatable XML strings.
	 * @param string   $entity_id  Related entity ID.
	 * @return string[]
	 */
	protected function send_translation_request( $from, $to, $xml_values, $entity_id = null ) {

		if ( ! $this->is_authorized() ) {
			return array();
		}

		// prepare translation request.
		$id         = uniqid();
		$xml_prefix = '<?xml version="1.0" encoding="utf-8" standalone="yes" ?><root>' . $this->segment_opening;
		$xml_suffix = $this->segment_closing . '</root>';
		$content    = $xml_prefix . implode( $this->delimiter, $xml_values ) . $xml_suffix;

		if ( strlen( $content ) === 0 ) {
			return array();
		}

		$post = $this->get_post_body( $id, $from, $to, $content, $entity_id );

		$options = array(
			'headers' => array(
				'Content-Type'   => 'application/json',
				'Content-Length' => strlen( $post ),
			),
			'body'    => $post,
			'auth'    => array( $this->get_application_name(), $this->get_password(), 'digest' ),
		);

		try {
			// send translation request.
			$response = $this->http_client->post( self::$api_url . '/translate', $options );
			// check request response.
			$body       = (string) $response->getBody();
			$request_id = is_numeric( $body ) ? (int) $body : null;
			$status     = $response->getStatusCode();
			if ( 200 !== $status || $request_id < 0 ) {
				$message = self::$error_map[ $request_id ] ?? $body;
				\Drupal::logger( WEBT_LOGGER )->error( "Invalid request response from eTranslation: $response [status: $status, message: $message]" );
				return array();
			}

			\Drupal::logger( WEBT_LOGGER )->info( "eTranslation request successful ($request_id)" );

			// wait for translation callback.
			$translations = $this->await_etranslation_response( $request_id, null === $entity_id );
			if ( $translations ) {
				return $this->etranslation_response_to_translation_array( $translations->data );
			} else {
				return array();
			}
		} catch ( \GuzzleHttp\Exception\RequestException $e ) {
			$response = $e->getResponse();
			$message  = $response ? (string) $response->getBody() : '';
			if ( $response && strlen( $message ) == 0 ) {
				$message = \GuzzleHttp\Psr7\Message::toString( $response );
			}
			\Drupal::logger( WEBT_LOGGER )->error( 'Error translating strings: ' . $message );
			return array();
		} catch ( \GuzzleHttp\Exception\ConnectException $e ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'Could not establish connection with eTranslation. Error: ' . $e->getMessage() );
			return array();
		}
	}

	/**
	 * Split translated string response to separate XML strings.
	 *
	 * @param string $response Response body.
	 * @return string[]
	 */
	public function etranslation_response_to_translation_array( $response ) {
		// extract translations.
		$segment_tag_length = strlen( $this->segment_opening );
		$first_segment_pos  = strpos( $response, $this->segment_opening );
		$data_start_pos     = $first_segment_pos + $segment_tag_length;
		$data_end_pos       = strrpos( $response, $this->segment_closing );
		$segment_string     = substr( $response, $data_start_pos, $data_end_pos - $data_start_pos );

		$translations = explode( $this->delimiter, $segment_string );
		return $translations;
	}

	/**
	 * Builds translation request body.
	 *
	 * @param string $id Translation request ID.
	 * @param string $lang_from Source language.
	 * @param string $lang_to Target language.
	 * @param string $translatable_string Translatable text.
	 * @param string $entity_id Related entity ID.
	 * @return string
	 */
	private function get_post_body( $id, $lang_from, $lang_to, $translatable_string, $entity_id = null ) {
		$site_url = \Drupal::request()->getSchemeAndHttpHost();
		$document = array(
			'content'  => base64_encode( $translatable_string ),
			'format'   => 'xml',
			'filename' => 'translateMe',
		);

		$domain = $this->config->get( "mt_domain_$lang_to" );

		$translation_request_body = array(
			'documentToTranslateBase64' => $document,
			'sourceLanguage'            => strtoupper( $lang_from ),
			'targetLanguages'           => array(
				strtoupper( $lang_to ),
			),
			'errorCallback'             => $site_url . '/etranslation/error/' . $id,
			'callerInformation'         => array(
				'application' => $this->get_application_name(),
			),
			'destinations'              => array(
				'httpDestinations' => array(
					$site_url . '/etranslation/translation/' . $id,
				),
			),
			'domain'                    => $domain ? $domain : 'GEN',
		);
		if ( $entity_id ) {
			$translation_request_body['externalReference'] = $entity_id;
		}
		return json_encode( $translation_request_body, JSON_UNESCAPED_SLASHES );
	}

	/**
	 * Wait for async translation response and return its body.
	 *
	 * @param string  $request_id Request ID.
	 * @param boolean $is_pretranslation Whether this request is part of pre-translation process.
	 * @return string
	 */
	private function await_etranslation_response( $request_id, $is_pretranslation = true ) {
		$response          = null;
		$key               = EtranslationUtils::get_cache_key( $request_id );
		$start_time        = microtime( true );
		$check_interval_ms = 250;

		$timeout = $is_pretranslation ? $this->request_timeout : $this->etranslation_timeout;

		while ( ! $response && $timeout > 0 && microtime( true ) - $start_time < $timeout ) {
			$response = \Drupal::cache()->get( $key );
			usleep( $check_interval_ms * 1000 );
		}
		if ( ! $response ) {
			\Drupal::logger( WEBT_LOGGER )->info( "eTranslation response timeout ($request_id, $timeout s)" );
			\Drupal::cache()->set( $key, EtranslationUtils::$timeout_value );
			return null;
		} else {
			\Drupal::cache()->delete( $key );
			if ( EtranslationUtils::$error_value === $response->data ) {
				return null;
			}
			return $response;
		}
	}

	/**
	 * Send language direction (domain) request.
	 *
	 * @return Psr\Http\Message\ResponseInterface
	 */
	protected function send_language_direction_request() {
		$options = array(
			'auth' => array( $this->get_application_name(), $this->get_password(), 'digest' ),
		);

		try {
			return $this->http_client->get( self::$api_url . '/get-domains', $options );
		} catch ( \GuzzleHttp\Exception\RequestException $e ) {
			$response = $e->getResponse();
			$message  = $response ? (string) $response->getBody() : '';
			if ( $response && strlen( $message ) == 0 ) {
				$message = \GuzzleHttp\Psr7\Message::toString( $response );
			}
			\Drupal::logger( WEBT_LOGGER )->error( 'Error on language direction request: ' . $message );
			return $response;
		} catch ( \GuzzleHttp\Exception\ConnectException $e ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'Could not establish connection with eTranslation. Error: ' . $e->getMessage() );
			return new \GuzzleHttp\Psr7\Response( 408 );
		}
	}

	/**
	 * Checks if MT provider data is configured.
	 *
	 * @return boolean
	 */
	public function is_authorized() {
		if ( ! $this->get_application_name() || ! $this->get_password() ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'eTranslation credentials not configured. Please go to Translation provider settings and enter MT authorization data!' );
			return false;
		}
		return true;
	}

	/**
	 * Map language direction request response to Generic API (Custom provider) JSON response.
	 *
	 * @param string $body Response body.
	 * @return string
	 */
	protected function map_language_direction_response_body( $body ) {
		$systems = array();
		foreach ( json_decode( $body ) as $domain => $value ) {
			$lang_pairs = $value->languagePairs;
			foreach ( $lang_pairs as $pair ) {
				$lower = strtolower( $pair );
				$langs = explode( '-', $lower );

				$system          = new \stdClass();
				$system->srcLang = $langs[0];
				$system->trgLang = $langs[1];
				$system->domain  = $domain;
				$system->name    = $value->name;
				$systems[]       = $system;
			}
		}
		$response                     = new \stdClass();
		$response->languageDirections = $systems;
		return json_encode( $response );
	}

	/**
	 * Retrieves user specified eTranslation application name from configuration.
	 *
	 * @return string
	 */
	private function get_application_name() {
		return $this->config->get( 'etranslation_appname' );
	}

	/**
	 * Retrieves and decrypts user specified eTranslation password from configuration.
	 *
	 * @return string or null if password not found
	 */
	private function get_password() {
		$encrypted = $this->config->get( 'etranslation_password' );
		if ( $encrypted ) {
			return EtranslationUtils::decrypt_password( $encrypted );
		}
		return null;
	}
}
