<?php

namespace Drupal\webt\translation_engines\generic;

use Drupal\webt\translation_engines\AbstractTranslationService;
use GuzzleHttp\Psr7;

/**
 * Machine translation service implementation for Custom provider.
 */
class GenericMTService extends AbstractTranslationService {

	/**
	 * Char limit which determines when to split translatable strings into separate request
	 *
	 * @var integer
	 */
	public static $default_chars_per_request = 1500;

	/**
	 * Set service parameters from config and initialize service.
	 *
	 * @param \Drupal\Core\Http\ClientFactory $http_client_factory HTTP client factory.
	 */
	public function __construct( \Drupal\Core\Http\ClientFactory $http_client_factory ) {
		parent::__construct( $http_client_factory );
		$this->max_chars_per_request = $this->config->get( 'mt_chars_per_request' ) ? $this->config->get( 'mt_chars_per_request' ) : self::$default_chars_per_request;
	}

	/**
	 * Translate given XML strings via Custom provider.
	 *
	 * @param string   $from       Source langcode.
	 * @param string   $to         Target langcode.
	 * @param string[] $xml_values Translatable XML strings.
	 * @param string   $entity_id  Related entity ID.
	 * @return string[]
	 */
	protected function send_translation_request( $from, $to, $xml_values, $entity_id = null ) {
		$data          = (object) array();
		$data->srcLang = $from;
		$data->trgLang = $to;
		$data->text    = $xml_values;

		$api_url = $this->config->get( 'mt_api_url' );
		$api_key = $this->config->get( 'mt_api_key' );
		$domain  = $this->config->get( "mt_domain_$to" );
		if ( $domain ) {
			$data->domain = $domain;
		}

		if ( ! $this->is_authorized() ) {
			return array();
		}

		$options = array(
			'headers' => array(
				'Content-Type' => 'application/json',
				'accept'       => 'text/plain',
				'X-API-KEY'    => $api_key,
			),
			'body'    => json_encode( $data, JSON_UNESCAPED_SLASHES ),
		);
		try {
			$response = $this->http_client->post( $api_url . '/translate/text', $options );
			$json     = json_decode( $response->getBody() );

			$translations = array_map(
				function( $item ) {
					return stripslashes( $item->translation );
				},
				$json->translations
			);
			return $translations;
		} catch ( \GuzzleHttp\Exception\RequestException $e ) {
			$response = $e->getResponse();
			$message  = $response ? (string) $response->getBody() : '';
			if ( $response && strlen( $message ) == 0 ) {
				$message = \GuzzleHttp\Psr7\Message::toString( $response );
			}
			\Drupal::logger( WEBT_LOGGER )->error( 'Error translating strings: ' . $message );
			return array();
		} catch ( \GuzzleHttp\Exception\ConnectException $e ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'Could not establish connection with the MT provider. Check your API URL or try again later! Error: ' . $e->getMessage() );
			return array();
		}
	}

	/**
	 * Send language direction (domain) request.
	 *
	 * @return Psr\Http\Message\ResponseInterface
	 */
	protected function send_language_direction_request() {
		$api_url = $this->config->get( 'mt_api_url' );
		$api_key = $this->config->get( 'mt_api_key' );

		$options = array(
			'headers' => array(
				'accept'    => 'text/plain',
				'X-API-KEY' => $api_key,
			),
		);

		try {
			return $this->http_client->get( $api_url . '/translate/language-directions', $options );
		} catch ( \GuzzleHttp\Exception\RequestException $e ) {
			$response = $e->getResponse();
			$message  = $response ? (string) $response->getBody() : '';
			if ( $response && strlen( $message ) == 0 ) {
				$message = \GuzzleHttp\Psr7\Message::toString( $response );
			}
			\Drupal::logger( WEBT_LOGGER )->error( 'Error on language direction request: ' . $message );
			return $response;
		} catch ( \GuzzleHttp\Exception\ConnectException $e ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'Could not establish connection with the MT provider. Check your API URL or try again later! Error: ' . $e->getMessage() );
			return new \GuzzleHttp\Psr7\Response( 408 );
		}
	}

	/**
	 * Checks if MT API URL & key is provided.
	 *
	 * @return boolean
	 */
	public function is_authorized() {
		$api_url = $this->config->get( 'mt_api_url' );
		$api_key = $this->config->get( 'mt_api_key' );
		if ( ! $api_url || ! $api_key ) {
			\Drupal::logger( WEBT_LOGGER )->error( 'MT API URL or key not set. Please go to Translation provider settings and enter MT authorization data!' );
			return false;
		}
		return true;
	}
}
